from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel, validator

from .exceptions import UnknownAction, UnknownGuard


def ALWAYS(*_: Any) -> bool:
    return True


def load_actions(values: List[Union[str, Callable]], funcs: Dict[str, Callable]) -> List[Callable]:
    """Load all of the actions in the passed values

    :param values: values to load
    :type values: List[Union[str, Callable]]
    :param funcs: functions to load from
    :type funcs: Dict[str, Callable]
    :raises UnknownAction: when an action can be looked up in the passed in actions
    :return: list of loaded callable actions
    :rtype: List[Callable]
    """
    loaded_actions = []
    for value in values:
        if isinstance(value, str):
            action = funcs.get(value)
            if action is not None:
                loaded_actions.append(action)
            else:
                raise UnknownAction(f"Unknown action '{value} - expected one of: {', '.join(funcs.keys())}'")
        else:
            loaded_actions.append(value)

    return loaded_actions


class Transition(BaseModel):
    """Represents a state transition in a state machine

    :param target: The target state to transition to.
    :type target: str
    :param cond: A function or string that represents the condition that must be met before the transition can occur.
                 If it is a string, it is assumed to be the name of a function that must be available in the current context.
    :type cond: Callable or str
    :param actions: A list of functions or strings that represent the actions to be executed during the transition.
                   If an item is a string, it is assumed to be the name of a function
                   that must be available in the current context.
    :type actions: List[Union[Callable, str]]
    """

    target: Optional[str]
    """The target state for this transition if the condition is met"""
    guards: Optional[Dict[str, Callable]]
    """Possible guards, which are callables"""
    actions_map: Optional[Dict[str, Callable]] = None
    """Action map"""
    actions: List[Callable] = []
    """Action side effects for the machine"""
    cond: Callable = ALWAYS
    """The condition for the transition"""

    @validator("actions", pre=True)
    def build_actions(cls: "Transition", value: Union[Callable, List], values: Dict[str, Any]) -> List[Callable]:
        """Builds the list of actions to be executed during the transition.

        :param value: A function or a list of functions that represent the actions to be executed during the transition.
        :type value: Union[Callable, List]
        :return: A list of functions that represent the actions to be executed during the transition.
        :rtype: List[Callable]
        """
        actions = values.get("actions_map")

        if callable(value):  # a single action value return as a list
            return [value]
        elif isinstance(value, str):  # string callable look it up!
            return load_actions([value], actions)
        elif actions:
            return load_actions(value, actions)  # load the list of actions

        return []

    @validator("cond", pre=True)
    def build_cond(cls: "Transition", value: Union[Callable, str], values: Dict[str, Any]) -> Optional[Callable]:
        """Builds the cond value for a transition

        :param value: the condition callable / str
        :type value: Union[Callable, str]
        :param values: values
        :type values: Dict[str, Any]
        :return: Callable
        :rtype: Callable
        """
        guards = values.get("guards")
        if callable(value):
            return value
        elif isinstance(value, str) and guards is not None:
            guard = guards.get(value)
            if guard is None:
                raise UnknownGuard(f"Unknown action '{value} - expected one of: {', '.join(guards.keys())}'")
            elif callable(guard):
                return guard  # type: ignore[no-any-return]
            else:
                return ALWAYS

        return ALWAYS



from typing import Tuple


def parse_target(target: str) -> Tuple[str, str]:
    """Parses a target path, providing the first item & remaining path

    :param target: The target for the transition
    :returns: Tuple[str, str]
    """
    target_state = None
    remainder = None
    while not target_state:
        values = target.split(".", maxsplit=1)
        target_state = values[0]
        target = remainder = values[-1]

    return target_state, remainder



from asyncio import iscoroutinefunction
from typing import Callable, Dict, List, Optional, Union

from pydantic import BaseModel, validator

from .transition import Transition


class Event(BaseModel):
    name: Optional[str] = None
    """Name of the event"""
    guards: Optional[Dict[str, Callable]]
    """Possible guards, which are callables"""
    actions: Optional[Dict[str, Callable]]
    """Action side effects for the machine"""
    transitions: List[Transition]

    @validator("transitions", pre=True)
    def build_transitions(cls, value: Union[Dict, List], values: Dict) -> List[Transition]:
        """Build a list of transitions from the given transition data

        :param transition_data: The transition data, either a single transition dict or a list of transition dicts
        :type transition_data: Union[Dict, List]

        :returns: The list of transitions
        :rtype: List[Transition]"""
        guards = values.get("guards")
        actions = values.get("actions")
        if isinstance(value, dict):
            return [Transition(guards=guards, actions_map=actions, **value)]
        else:
            return [Transition(guards=guards, actions_map=actions, **transition) for transition in value]

    async def execute_actions(self, context: dict, event: str, transition: Transition) -> None:
        """Execute all of the actions in a transition

        :param context: The context of the system
        :type context: dict

        :param event: The event that triggered the transition
        :type event: str

        :param transition: The transition that will be executed
        :type transition: Transition"""
        for action in transition.actions:
            if iscoroutinefunction(action):
                await action(context, event)
            else:
                action(context, event)

    async def get_transition(self, context: dict, event: str) -> Optional[Transition]:
        """Return the next transition to take, if a condition is met

        :param context: The context of the system
        :type context: dict

        :param event: The event that triggered the transition
        :type event: str

        :returns: The next transition to take, if a condition is met
        :rtype: Optional[Transition]
        """
        for transition in self.transitions:
            if transition.cond(context, event):
                await self.execute_actions(context=context, event=event, transition=transition)
                return transition

        return None



from logging import Logger, getLogger
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel, root_validator, validator

from .event import Event
from .exceptions import UnknownTarget
from .state import State
from .utils import parse_target


class Machine(BaseModel):
    name: str
    """The name for the machine"""
    initial: str
    """The name of the initial state"""
    state: Optional[State]
    """The current state for the machine"""
    guards: Optional[Dict[str, Callable]]
    """Possible guards, which are callables"""
    actions: Optional[Dict[str, Callable]]
    """Action side effects for the machine"""
    context: Optional[Dict[str, Any]]
    """The contextual information """
    events: Optional[Union[Dict[str, Event], List[Event]]] = {}
    """The events at the root of the machine"""
    states: Dict[str, State]
    """The possible states for the machine"""
    logger: Logger = getLogger(__name__)
    """The logger for the machine"""

    class Config:
        arbitrary_types_allowed = True

    @validator("states", pre=True)
    def build_states(cls, value: dict, values: Dict[str, Any]) -> Dict[str, State]:
        """Builds the states from the passed in states object.

        :param value: The state configuration dictionary.
        :type value: dict

        :return: The built state objects.
        :rtype: Dict[str, State]
        """
        guards = values.get("guards")
        actions = values.get("actions")
        return {name: State(name=name, actions=actions, guards=guards, **val) for name, val in value.items()}

    @root_validator
    def build_initial_state(cls, values: Dict[str, Any]) -> Dict:
        """Builds the initial state value.

        :param values: The values passed to the Machine constructor.
        :type values: Dict

        :return: The updated values.
        :rtype: Dict
        """
        if states := values.get("states"):
            initial = values["initial"]
            values["state"] = states.get(initial)
        # check for transient state and transition?
        return values

    @classmethod
    async def create(cls: "Machine", config: dict) -> "Machine":  # type: ignore[misc]
        """Creates a new instance of the Machine class.

        :param config: The machine configuration dictionary.
        :type config: dict

        :return: The created Machine instance.
        :rtype: Machine
        """
        machine = cls(**config)  # type: ignore[operator]
        await machine.step()  # make sure all transient states are executed for initial state
        return machine  # type: ignore[no-any-return]

    @validator("events", pre=True)
    def build_events(cls, value: dict) -> Dict[str, Event]:
        """Build the events

        :param value: The events to be built
        :type value: dict

        :returns: The built events as a dictionary
        :rtype: dict
        """
        return {name: Event(name=name, transitions=val) for name, val in value.items()}

    def update_config(self, config: dict) -> None:
        """Updates this instances config with the passed in config.

        :param config: The configuration to update the instance with.
        :type config: dict
        """
        return None

    async def event(self, event: str) -> State:
        """Transitions the machine by executing an event

        :param event: The name of the event to trigger
        :type event: str

        :returns: The current state of the machine after the transition
        :rtype: State
        """
        self.logger.info("Machine processing event: %s", event)
        event = self.events.get(event) if event in self.events else self.state.get_event(event)  # type: ignore[union-attr, assignment, operator]
        if event:
            transition = await event.get_transition(self.context, event)  # type: ignore[attr-defined]
            if transition:  # there is a transition
                if transition.target is not None:
                    await self.do_transition(target=transition.target)  # if the transition has a target set the state
                await self.step()  # step through the machine as state changed
        else:
            self.logger.error("Event %s not found", event)

        return self.state  # type: ignore[return-value]

    async def step(self, state: Optional[State] = None) -> State:
        """Step through the machine until no more transitions to move through

        :param state: The state to start the step process from, defaults to None which will use the current state of the machine
        :type state: Optional[State], optional

        :returns: The final state after all possible transitions have been processed
        :rtype: State
        """
        if state is None:
            state = self.state

        transition = await self.state.get_transition(context=self.context)  # type: ignore[union-attr , arg-type]
        if transition:
            await self.do_transition(transition.target)
            return await self.step()  # step through the new state

        return state  # type: ignore[return-value]

    async def on_entry(self, context: dict) -> None:
        """Perform any entry actions for the new state

        :param context: The context data for the current state
        :type context: dict
        """
        await self.state.on_entry(context)  # type: ignore[union-attr]

    async def on_exit(self, context: dict) -> None:
        """Perform any exit actions for the current state

        :param context: The context data for the current state
        :type context: dict
        """
        await self.state.on_exit(context)  # type: ignore[union-attr]

    async def update_state(self, target: str) -> bool:
        """Update the current state to the target state

        :raises: UnknownTarget - If the target state can not be found
        """
        target_state, remainder = parse_target(target=target)
        state = self.states.get(target_state)
        if state is None:
            # if the state is target state is None pass to child state to handle
            if await self.state.update_state(target, self.context) is None:  # type: ignore[union-attr, arg-type]
                raise UnknownTarget("Target state can not be found")
            return False
        else:
            if self.state:
                await self.on_exit(self.context)  # type: ignore[arg-type]
            self.state = state
            if remainder and remainder != target_state:
                # consume the rest of the path!
                await self.state.update_state(remainder, self.context)  # type: ignore[arg-type]
            return True

    async def do_transition(self, target: str) -> None:
        """Set a state from a target

        :param target: The name of the state to transition to
        :type target: str
        """
        if await self.update_state(target):
            await self.on_entry(self.context)  # type: ignore[arg-type]

        return None

    async def with_context(self, context: dict) -> State:
        """Update the context and step through the machine

        :param context: The context to update the machine with
        :type context: dict

        :returns: The final state after all possible transitions have been processed
        :rtype: State
        """
        self.context = context  # update the context
        return await self.step()  # step through the machine

    @property
    def initial_state(self) -> State:
        """Get the initial state of the machine

        :returns: The initial state of the machine
        :rtype: State
        """
        return self.states[self.initial]



class UnknownTarget(Exception):
    """Target state can not be found"""


class UnknownAction(Exception):
    """Action could not be found"""


class UnknownGuard(Exception):
    """Guard could not be found"""



import asyncio
from enum import Enum
from logging import Logger, getLogger
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel, root_validator, validator

from .event import Event
from .transition import Transition
from .utils import parse_target

ALWAYS = "always"


class StateType(str, Enum):
    atomic = "atomic"
    compound = "compound"
    final = "final"
    transient = "transient"


class State(BaseModel):
    type: Optional[StateType]  # noqa: A003
    """The type of state this is"""
    name: str
    """The name for this state"""
    state: Optional["State"] = None
    """The sub-state for this state"""
    entry: List[Callable] = []
    """Callables to call when entering this state"""
    exit: List[Callable] = []  # noqa: A003
    """Callables to call when exiting this state"""
    parent: Optional[str] = "."
    """The parent of this state"""
    initial: Optional[str] = None
    """The name of the initial sub-state for this state"""
    guards: Optional[Dict[str, Callable]]
    """Possible guards, which are callables"""
    actions: Optional[Dict[str, Callable]]
    """Action side effects for the machine"""
    transitions: List[Transition] = []
    """Transient transitions to call when stepping through the state"""
    states: Optional[Dict[str, "State"]] = {}
    """All possible state for this state"""
    events: Optional[Union[Dict[str, Event], List[Event]]] = {}
    """Possible events to execute for this state"""
    logger: Logger = getLogger(__name__)

    class Config:
        arbitrary_types_allowed = True

    @validator("states", pre=True)
    def build_states(cls, value: dict, values: dict) -> Dict[str, "State"]:
        """Build the states from the passed in states object

        :param value: The states object to be built
        :type value: dict

        :returns: The built states as a dictionary
        :rtype: dict
        """
        current_name = values.get("name")
        parent_name = values.get("parent")
        guards = values.get("guards")
        actions = values.get("actions")
        formatted_parent = f"{parent_name}{'.' if parent_name != '.' else ''}{current_name}"

        return {
            name: State(
                name=name,
                guards=guards,
                actions=actions,
                parent=formatted_parent,
                **val,
            )
            for name, val in value.items()
        }

    @root_validator
    def build_type(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Builds the type based on the values passed in

        - Compound is defined as having sub-states.
        - Atomic is defined as having events but no sub-states
        - Transient is defined as only having transitions
        - Final is defined as having no further state changes
        """
        states = values.get("states")
        transitions = values.get("transitions")
        events = values.get("events")
        _type = None
        if states:
            _type = StateType.compound
        elif events:
            _type = StateType.atomic
        elif transitions:
            _type = StateType.transient
        else:
            _type = StateType.final

        values["type"] = _type
        return values

    @validator("events", pre=True)
    def build_events(cls, value: dict, values: Dict[str, Any]) -> Dict[str, Event]:
        """Build the events

        :param value: The events to be built
        :type value: dict

        :returns: The built events as a dictionary
        :rtype: dict
        """
        guards = values.get("guards")
        actions = values.get("actions")
        return {name: Event(name=name, guards=guards, actions=actions, transitions=val) for name, val in value.items()}

    @validator("transitions", pre=True)
    def build_transitions(cls, value: Optional[list], values: dict) -> List[Transition]:
        """Build the always transient transitions

        :param value: The transitions to be built
        :type value: Optional[list]

        :returns: The built transitions as a list
        :rtype: list
        """
        guards = values.get("guards")
        actions = values.get("actions")
        if isinstance(value, dict):
            return [Transition(target=value["target"], guards=guards, actions_map=actions)]
        else:
            return [Transition(guards=guards, actions_map=actions, **transition) for transition in value]  # type: ignore[union-attr]

    @validator("entry", pre=True)
    def build_entry(cls, value: Optional[list]) -> List[Callable]:
        """Setup the entry callables

        :param value: The callables to be set up
        :type value: Optional[list]

        :returns: The set up callables as a list
        :rtype: list
        """
        return [value] if callable(value) else value  # type: ignore[return-value, list-item]

    @validator("exit", pre=True)
    def build_exit(cls, value: Optional[list]) -> List[Callable]:
        """Setup the exit callables

        :param value: The callables to be set up
        :type value: Optional[list]

        :returns: The set up callables as a list
        :rtype: list
        """
        return [value] if callable(value) else value  # type: ignore[return-value, list-item]

    @property
    def value(self) -> Union[str, Dict[str, Any]]:
        """Return the value of the state

        :returns: The name of the state
        :rtype: str
        """
        return {self.name: self.state.value} if self.state else self.name

    def get_event(self, name: str) -> Optional[Event]:
        """Return event for the name

        :param name: The name of the event to return
        :type name: str

        :returns: The event with the given name
        :rtype: Event
        """
        self.logger.info("State %s finding event %s", self.name, name)
        event = self.events.get(name)  # type: ignore[union-attr]
        if event is None and self.state:  # this state does not have the event!
            # check the current child state of for this event
            self.logger.info("%s found event %s", self.name, name)
            event = self.state.get_event(name=name)
        else:
            self.logger.info("State %s found event %s", self.name, name)
        return event

    async def on_entry(self, context: dict) -> None:
        """Called when the state is entered

            - Handle compound states
            - Call `on_entry`

        :param context: A dictionary that contains the current context of the system
        :type context: dict
        """
        self.logger.info("Entering %s state in %s parent", self.name, self.parent)
        await self._process_callables(self.entry, context)  # state entered so process entry handlers
        if self.states and not self.state:  # if there sub-states they should be processed!
            self.state = self.states.get(str(self.initial))  # set the initial state

            # call the state's on_entry to setup any further sub-states
            await self.state.on_entry(context=context)  # type: ignore[union-attr]

    async def on_exit(self, context: dict) -> None:
        """Called when a state is exited

        :param context: A dictionary that contains the current context of the system
        :type context: dict
        """
        self.logger.info("Exiting %s state in %s parent", self.name, self.parent)
        await self._process_callables(self.exit, context)
        if self.state:
            await self.state.on_exit(context=context)  # call the sub-state's on_exit to exit any sub-states
            self.state = None  # reset the stored state upon exiting

    async def _process_callables(self, callables: List[Callable], context: dict) -> None:
        """Process the given callables with the given context

        :param context: A dictionary that contains the current context of the system
        :type context: dict
        :param callables: A dictionary that contains the current context of the system
        :type callables: List[Callable]
        """
        for func in callables:
            if asyncio.iscoroutinefunction(func):
                await func(context)
            else:
                func(context)

    async def update_state(self, target: str, context: dict) -> Optional["State"]:
        """Update the current state to the target state

        :raises: UnknownTarget - If the target state can not be found
        """
        target_state, remainder = parse_target(target=target)

        state = self.states.get(target_state)  # type: ignore[union-attr]
        if state is None:  # noqa: SIM102
            # if there is a sub-state then pass this onto the next state
            if self.state:
                return await self.state.update_state(target=target, context=context)
        else:
            if self.state:  # exiting state
                await self.state.on_exit(context=context)
            self.state = state
            await self.state.on_entry(context)
            if remainder and remainder != target_state:
                await self.state.update_state(target=remainder, context=context)
            return state

        return None

    async def get_transition(self, context: dict) -> Optional[Transition]:
        """Check the conditions for each transition of the given state

        :param context: A dictionary that contains the current context of the system
        :type context: dict

        :returns: The next transition to take, if a condition is met
        :rtype: Optional[Transition]
        """

        if self.state:  # noqa: SIM102
            # Go depth first to the last state
            if new_transition := await self.state.get_transition(context=context):
                await self.on_exit(context=context)
                await self.update_state(target=new_transition.target, context=context)  # type: ignore[arg-type]

        for transition in self.transitions:
            if asyncio.iscoroutinefunction(transition.cond):
                condition = await transition.cond(context, None)
            else:
                condition = transition.cond(context, None)

            if condition:
                return transition

        return None
